﻿/*
    Copyright 2002-2014 KeyDays LLC.

    This file is a part of KeyDays.Lib.
    KeyDays.Lib is licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. 
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

using KeyDays.Lib;
using MyDiagno = KeyDays.Lib.Diagnostics;

namespace KeyDays.Lib.Drawing.Imaging
{
    #region helpers

    public class ImageProperties
    {
        #region fields
        public readonly long SizeInByte;
        #endregion fields

        #region properties
        public Size Dimensions { get; private set; }
        public float Size { get; private set; }
        public int MinQuality { get; private set; }
        #endregion properties

        #region constructors
        public ImageProperties(Size dimensions, float Size)
            : this(dimensions, Size, 100)
        {
            //--- Do nothing.
        }
        public ImageProperties(Size dimensions, float Size, int minquality)
        {
            this.Dimensions = dimensions;
            this.Size = Size;
            this.SizeInByte = (long)(this.Size * 1024.00f);
            this.MinQuality = minquality;
        }
        #endregion constructors

        #region metods
        public override string ToString()
        {
            return string.Format("Dimensions({0}, {1}), Size={2}, MinQuality={3}", this.Dimensions.Width, this.Dimensions.Height, this.Size, this.MinQuality);
        }
        #endregion metods
    }

    public sealed class ImageExt
    {
        public readonly string FileName;
        public readonly ImageCodecInfo CodecInfo;
        public Image Image;
        public long SizeInByte;

        public ImageExt(string fileName)
        {
            this.FileName = fileName;
            //-- http://stackoverflow.com/questions/788335/why-does-image-fromfile-keep-a-file-handle-open-sometimes
            //this.Image = Image.FromFile(this.FileName);
            MemoryStream objMemoryStream = new MemoryStream(File.ReadAllBytes(this.FileName));
            this.Image = Image.FromStream(objMemoryStream);
            this.SizeInByte = objMemoryStream.Length;

            string
                FileExtension = Path.GetExtension(this.FileName).ToLower(),
                MimeType = null;

            this.CodecInfo = null;
            if (false == ImageManager.FileExtension2MimeType.Keys.Contains(FileExtension))
            {
                MimeType = ImageManager.FileExtension2MimeType[FileExtension];
                if (1 == ImageCodecInfo.GetImageEncoders().Count(rec => rec.MimeType == MimeType))
                {
                    this.CodecInfo = ImageCodecInfo.GetImageEncoders().First(rec => rec.MimeType == MimeType);
                }
            }
            if (null == this.CodecInfo)
            {
                MimeType = ImageManager.FileExtension2MimeType[ImageManager.FILEEXTENSION];
                //--- Remember at the begining we made sure that we have ImageCodecInfo for the default file extention.
                this.CodecInfo = ImageCodecInfo.GetImageEncoders().First(rec => rec.MimeType == MimeType);
            }

            // this.SizeInByte = ImageExt.GetImageSizeInByte(this.Image);
        }

        public static long GetImageSizeInByte(Image image)
        {
            MemoryStream objMemoryStream = new MemoryStream();
            image.Save(objMemoryStream, ImageFormat.Jpeg);
            objMemoryStream.Flush();
            long ImageFileSize = objMemoryStream.Length;
            objMemoryStream.Close();
            return ImageFileSize;
        }
    }

    public class InvalidImageFile : ApplicationException
    {
        public readonly FileInfo SourcefileFileInfo;
        public InvalidImageFile(FileInfo fileInfo)
            : base(string.Format("Invalid image file: {0}", fileInfo.FullName))
        {
            this.SourcefileFileInfo = fileInfo;
        }
    }
    #endregion helpers

    /// <summary>
    /// 
    /// </summary>
    /// <remarks>This class is not thread safe.</remarks>
    public class ImageManager
    {
        #region heper
        internal sealed class ImageResult
        {
            internal long
                _ImageSizeInByte,
                _Difference;

            internal int _Quality;

            internal Image _Image;
        }
        public enum ImageSizeAdjustmentTypes
        {
            Auto = 0,
            Width = 1,
            Height = 2,
        }
        #endregion heper

        #region fileds
        public static readonly Size MAXSIZE;
        public static readonly int QUALITY;
        public static readonly bool STRECH;
        public static readonly string FILEEXTENSION;
        public static readonly Dictionary<string, string> FileExtension2MimeType;
        #endregion fileds

        #region properties
        public Size MaxSize { get; set; }
        public int Quality { get; set; }
        public bool Strech { get; set; }
        public string FileExtention { get; set; }
        public const ulong FileSizeLT = 0x0001; //0000 0000 0000 0001
        public const ulong FileSizeGT = 0x0002; //0000 0000 0000 0010
        public const ulong ImageHeightLT = 0x0001; //0000 0000 0000 0001
        public const ulong ImageHeightGT = 0x0002; //0000 0000 0000 0010
        public const ulong ImageWidthLT = 0x0001; //0000 0000 0000 0001
        public const ulong ImageWidthGT = 0x0002; //0000 0000 0000 0010
        #endregion properties

        #region constructors
        static ImageManager()
        {
            //ImageManager.MAXSIZE = new Size(320, 240);
            ImageManager.MAXSIZE = new Size(800, 600);
            ImageManager.QUALITY = 100;
            ImageManager.STRECH = false;
            ImageManager.FILEEXTENSION = ".jpg";
            ImageManager.FileExtension2MimeType = new Dictionary<string, string>();

            ImageManager.FileExtension2MimeType.Add(".bmp", "image/bmp");
            ImageManager.FileExtension2MimeType.Add(".jpg", "image/jpeg");
            ImageManager.FileExtension2MimeType.Add(".jpeg", "image/jpeg");
            ImageManager.FileExtension2MimeType.Add(".jpe", "image/jpeg");
            ImageManager.FileExtension2MimeType.Add(".png", "image/png");
            ImageManager.FileExtension2MimeType.Add(".gif", "image/gif");
            ImageManager.FileExtension2MimeType.Add(".ico", "image/x-icon");
            ImageManager.FileExtension2MimeType.Add(".tif", "image/tiff");
            ImageManager.FileExtension2MimeType.Add(".tiff", "image/tiff");
            ImageManager.FileExtension2MimeType.Add(".cod", "image/cis-cod");
            ImageManager.FileExtension2MimeType.Add(".ief", "image/ief");
            ImageManager.FileExtension2MimeType.Add(".jfif", "image/pipeg");
            ImageManager.FileExtension2MimeType.Add(".svg", "image/svg+xml");
            ImageManager.FileExtension2MimeType.Add(".ras", "image/x-cmu-raster");
            ImageManager.FileExtension2MimeType.Add(".cmx", "image/x-cmx");
            ImageManager.FileExtension2MimeType.Add(".pnm", "image/x-portable-anymap");
            ImageManager.FileExtension2MimeType.Add(".pbm", "image/x-portable-bitmap");
            ImageManager.FileExtension2MimeType.Add(".pgm", "image/x-portable-graymap");
            ImageManager.FileExtension2MimeType.Add(".ppm", "image/x-portable-pixmap");
            ImageManager.FileExtension2MimeType.Add(".rgb", "image/x-rgb");
            ImageManager.FileExtension2MimeType.Add(".xbm", "image/x-xbitmap");
            ImageManager.FileExtension2MimeType.Add(".xpm", "image/x-xpixmap");
            ImageManager.FileExtension2MimeType.Add(".xwd", "image/x-xwindowdump");

            //--- By accessingt these values we are making sure that we are in the position to process the default values later...
            string
                DefaultMimeType = ImageManager.FileExtension2MimeType[ImageManager.FILEEXTENSION];
            int ImageCodecInfoCount = ImageCodecInfo.GetImageEncoders().Count(rec => rec.MimeType == DefaultMimeType);
            System.Diagnostics.Trace.Assert(1 == ImageCodecInfoCount, string.Format("ImageCodecInfo not found for: (FILEEXTENSION={0}, DefaultMimeType={1})", ImageManager.FILEEXTENSION, DefaultMimeType));
        }
        public ImageManager()
            : this(ImageManager.MAXSIZE, ImageManager.QUALITY)
        {
            //--- Do nothing.
        }
        public ImageManager(ImageProperties imageProperties)
            : this(imageProperties.Dimensions, imageProperties.MinQuality)
        {

        }
        public ImageManager(Size maxSize, int minQuality)
        {
            this.MaxSize = maxSize;
            this.Quality = minQuality;
            this.Strech = ImageManager.STRECH;
            this.FileExtention = ImageManager.FILEEXTENSION;
        }
        #endregion constructors

        #region methods
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName)
        {
            this.ScaleDownImages(sourceImageFolderName, tragetImageFolderName, this.MaxSize);
        }
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName, Size maxSize)
        {
            this.ScaleDownImages(sourceImageFolderName, tragetImageFolderName, maxSize, this.Quality);
        }
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName, Size maxSize, int quality)
        {
            this.ScaleDownImages(sourceImageFolderName, tragetImageFolderName, maxSize, quality, string.Format("*{0}", this.FileExtention));
        }
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName, Size maxSize, int quality, string searchPattern)
        {
            this.ScaleDownImages(sourceImageFolderName, tragetImageFolderName, maxSize, quality, searchPattern, ImageSizeAdjustmentTypes.Auto);
        }
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName, Size maxSize, int quality, string searchPattern, ImageSizeAdjustmentTypes imageSizeAdjustmentType)
        {
            this.ScaleDownImages(sourceImageFolderName, tragetImageFolderName, maxSize, quality, searchPattern, ImageSizeAdjustmentTypes.Auto, this.Strech);
        }
        public void ScaleDownImages(string sourceImageFolderName, string tragetImageFolderName, Size maxSize, int quality, string searchPattern, ImageSizeAdjustmentTypes imageSizeAdjustmentType, bool strech)
        {
            string[] FileNames = Directory.GetFiles(sourceImageFolderName, searchPattern, SearchOption.TopDirectoryOnly);
            if (0 == FileNames.Length)
            {
                return;
            }

            tragetImageFolderName = Path.GetFullPath(tragetImageFolderName);

            string BackupFolderName = null;

            if (true == Directory.Exists(tragetImageFolderName))
            {
                //--- If trager folder has sub folders then do not proceed.
                if (0 != Directory.GetDirectories(tragetImageFolderName).Count())
                {
                    throw new NotSupportedException(string.Format("Target folder {0} has sub folders. This case is not supported at this time.", tragetImageFolderName));
                }
                //--- To be safe, rename the target folder if it exists already...
                BackupFolderName = string.Format("{0}-[{1}].Backup", tragetImageFolderName, DateTime.Now.ToString("s").Replace(":", "-"));
                BackupFolderName = BackupFolderName.Replace(@"\-[", "-[");
                Directory.Move(tragetImageFolderName, BackupFolderName);
            }

            Directory.CreateDirectory(tragetImageFolderName);

            try
            {
                string
                    TargetFileName = null;
                foreach (string FileName in FileNames)
                {
                    TargetFileName = Path.Combine(tragetImageFolderName, Path.GetFileName(FileName));
                    this.ScaleDownImage(FileName, TargetFileName, maxSize, quality, imageSizeAdjustmentType, strech);
                }
            }
            catch (Exception exp)
            {
                MyDiagno.Trace.TraceError(exp, SeverityLevels.Medium);
                MyDiagno.Trace.TraceInformation("Beging: Rollback.");

                Directory.Delete(tragetImageFolderName, true);

                if (null != BackupFolderName)
                {
                    Directory.Move(BackupFolderName, tragetImageFolderName);
                }

                MyDiagno.Trace.TraceInformation("");
                MyDiagno.Trace.TraceInformation("Finished: Rollback.");

                throw;
            }
        }

        public void ScaleDownImage(string sourceImageFileName, string tragetImageFileName, Size maxSize)
        {
            this.ScaleDownImage(sourceImageFileName, tragetImageFileName, maxSize, this.Quality);
        }
        public void ScaleDownImage(string sourceImageFileName, string tragetImageFileName, Size maxSize, int quality)
        {
            this.ScaleDownImage(sourceImageFileName, tragetImageFileName, maxSize, quality, ImageSizeAdjustmentTypes.Auto, this.Strech);
        }

        //---------------------------
        public Image ScaleDownImage(ImageExt imageExt, Size maxSize, int quality, out long newImageSizeInByte)
        {
            return this.ScaleDownImage(imageExt, maxSize, quality, out newImageSizeInByte, ImageSizeAdjustmentTypes.Auto, this.Strech);
        }
        public Image ScaleDownImage(ImageExt imageExt, Size maxSize, int quality, out long newImageSizeInByte, ImageSizeAdjustmentTypes imageSizeAdjustmentType, bool strech)
        {
            return this.ScaleDownImage(imageExt.Image, maxSize, quality, out newImageSizeInByte, imageSizeAdjustmentType, strech, imageExt.CodecInfo);
        }
        //---------------------------
        public string ScaleDownImage(string sourceImageFileName, string tragetImageFileName, Size maxSize, int quality, ImageSizeAdjustmentTypes imageSizeAdjustmentType, bool strech)
        {
            string
                SourceImageFileExtension = Path.GetExtension(sourceImageFileName).ToLower(),
                TargetImageFileExtension = SourceImageFileExtension,
                TargetMimeType = null;

            ImageCodecInfo aImageCodecInfo = null;

            if (false == ImageManager.FileExtension2MimeType.Keys.Contains(SourceImageFileExtension))
            {
                TargetImageFileExtension = this.FileExtention;
            }

            TargetMimeType = ImageManager.FileExtension2MimeType[TargetImageFileExtension];

            if (1 == ImageCodecInfo.GetImageEncoders().Count(rec => rec.MimeType == TargetMimeType))
            {
                aImageCodecInfo = ImageCodecInfo.GetImageEncoders().First(rec => rec.MimeType == TargetMimeType);
            }
            else
            {
                //--- Remember at the begining we made sure that we have ImageCodecInfo for the default file extention.
                TargetImageFileExtension = this.FileExtention;
                aImageCodecInfo = ImageCodecInfo.GetImageEncoders().First(rec => rec.MimeType == TargetMimeType);
            }

            long NewImageSizeInByte = 0;
            Image
                InputImage = Image.FromFile(sourceImageFileName),
                OutputImage = this.ScaleDownImage(InputImage, maxSize, quality, out NewImageSizeInByte, imageSizeAdjustmentType, strech, aImageCodecInfo);

            if (TargetImageFileExtension != Path.GetExtension(tragetImageFileName))
            {
                tragetImageFileName = tragetImageFileName.Replace(Path.GetExtension(tragetImageFileName), TargetImageFileExtension);
            }

            OutputImage.Save(tragetImageFileName);

            OutputImage.Dispose();
            InputImage.Dispose();

            return tragetImageFileName;
        }
        public Image ScaleDownImage(Image sourceImage, Size maxSize, int quality, out long newImageSizeInByte, ImageSizeAdjustmentTypes imageSizeAdjustmentType, bool strech, ImageCodecInfo encoder)
        {
            // Image Image2Return = (Image) sourceImage.extParameterValidation("sourceImage").Clone();
            Size AdjustedSize = ImageManager.CalculateScaleDownImageSize(sourceImage.Size, maxSize, imageSizeAdjustmentType, strech);

            //--- If quality is ImageManager.QUALITY and the scaled down size does not exceed the orginal size then we are done.
            if (quality == this.Quality && AdjustedSize.Width >= sourceImage.Size.Width && AdjustedSize.Height >= sourceImage.Size.Height)
            {
                newImageSizeInByte = ImageExt.GetImageSizeInByte(sourceImage);
                return (Image)sourceImage.Clone();
            }

            Bitmap objBitmap = new Bitmap(sourceImage, AdjustedSize);
            EncoderParameters aEncoderParameters = new EncoderParameters(1);
            aEncoderParameters.Param[0] = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, quality);
            MemoryStream aMemoryStream = new MemoryStream();
            objBitmap.Save(aMemoryStream, encoder, aEncoderParameters);

            objBitmap.Dispose();
            objBitmap = null;

            newImageSizeInByte = aMemoryStream.Length;

            Image objImage = Image.FromStream(aMemoryStream);
            //aMemoryStream.Flush();
            //aMemoryStream.Close();
            //return (Image) objImage.Clone();
            return objImage;
        }
#if(OLD_CODE)
        public void ScaleDownImage2FitMaxFileSize(string imageFileName, long maxFileSizeInByte)
        {
            ImageExt objImageExt = new ImageExt(imageFileName);

            long
                Difference = objImageExt.SizeInByte - maxFileSizeInByte;

            if (0 >= Difference)
            {
                return;
            }

            int MinQuality = this.Quality,
                MidQuality = 0,
                MaxQuality = 99,
                OptimizedQuality = this.Quality,
                CacheOptimizedQuality = -1,
                CacheMinQuality = -1,
                CacheMidQuality = -1,
                CacheMaxQuality = -1;

            if (MaxQuality <= MinQuality)
            {
                throw new ApplicationException(string.Format("Minimum Quality ({0}) should be less than Maximum Qauality ({1})", this.Quality, MaxQuality));
            }

            int BestImage = -1;

            Image
                ImageTemp = null,
                ImageLo = null,
                ImageMd = null,
                ImageHi = null;

            long
                SizeInByteTemp = 0,
                SizeInByteLo = 0,
                SizeInByteMd = 0,
                SizeInByteTempMd = 0,
                SizeInByteHi = 0;
            try
            {
                while (MinQuality < MaxQuality)
                {
                    MidQuality = (MinQuality + MaxQuality) / 2;

                    if (CacheMidQuality != MidQuality)
                    {
                        if (null != ImageTemp) { ImageTemp.Dispose(); ImageTemp = null; }
                        CacheMidQuality = MidQuality;
                        ImageTemp = this.ScaleDownImage(objImageExt, this.MaxSize, MidQuality, out SizeInByteTemp);
                        SizeInByteTempMd = SizeInByteTemp; //  (SizeInByteTemp = ImageExt.GetImageSizeInByte(ImageTemp));
                    }
                    if (Difference > System.Math.Abs(SizeInByteTemp - maxFileSizeInByte))
                    {
                        if (null != ImageLo) { ImageLo.Dispose(); ImageLo = null; }
                        if (null != ImageMd) { ImageMd.Dispose(); ImageMd = null; }
                        if (null != ImageHi) { ImageHi.Dispose(); ImageHi = null; }

                        ImageMd = (Image)ImageTemp.Clone();
                        SizeInByteMd = SizeInByteTemp;
                        OptimizedQuality = MidQuality;
                        Difference = System.Math.Abs(SizeInByteMd - maxFileSizeInByte);
                        BestImage = 1;

                        if (0 == Difference / 1024)
                        {
                            break;
                        }
                    }

                    if (CacheMaxQuality != MaxQuality)
                    {
                        if (null != ImageTemp) { ImageTemp.Dispose(); ImageTemp = null; }
                        CacheMaxQuality = MaxQuality;
                        ImageTemp = this.ScaleDownImage(objImageExt, this.MaxSize, MaxQuality, out SizeInByteTemp);
                    }
                    if (Difference > System.Math.Abs(SizeInByteTemp - maxFileSizeInByte))
                    {
                        if (null != ImageLo) { ImageLo.Dispose(); ImageLo = null; }
                        if (null != ImageMd) { ImageMd.Dispose(); ImageMd = null; }
                        if (null != ImageHi) { ImageHi.Dispose(); ImageHi = null; }

                        ImageHi = (Image)ImageTemp.Clone();
                        SizeInByteHi = SizeInByteTemp;
                        OptimizedQuality = MaxQuality;
                        Difference = System.Math.Abs(SizeInByteHi - maxFileSizeInByte);
                        BestImage = 2;

                        if (0 == Difference / 1024)
                        {
                            break;
                        }
                    }

                    if (CacheMinQuality != MinQuality)
                    {
                        if (null != ImageTemp) { ImageTemp.Dispose(); ImageTemp = null; }
                        CacheMinQuality = MinQuality;
                        ImageTemp = this.ScaleDownImage(objImageExt, this.MaxSize, MinQuality, out SizeInByteTemp);
                    }
                    if (Difference > System.Math.Abs(SizeInByteTemp - maxFileSizeInByte))
                    {
                        if (null != ImageLo) { ImageLo.Dispose(); ImageLo = null; }
                        if (null != ImageMd) { ImageMd.Dispose(); ImageMd = null; }
                        if (null != ImageHi) { ImageHi.Dispose(); ImageHi = null; }

                        ImageLo = (Image)ImageTemp.Clone();
                        SizeInByteLo = SizeInByteTemp;
                        OptimizedQuality = MinQuality;
                        Difference = System.Math.Abs(SizeInByteLo - maxFileSizeInByte);
                        BestImage = 0;

                        if (0 == Difference / 1024)
                        {
                            break;
                        }
                    }

                    if (CacheOptimizedQuality == OptimizedQuality)
                    {
                        break;
                    }
                    CacheOptimizedQuality = OptimizedQuality;

                    if (SizeInByteTempMd > maxFileSizeInByte)
                    {
                        MaxQuality = MidQuality - 1;
                        MinQuality++;
                    }
                    else
                    {
                        MaxQuality--;
                        MinQuality = MidQuality + 1;
                    }
                }
            }
            catch (Exception exp)
            {
                MyDiagno.Trace.TraceError(exp, SeverityLevels.High);
                throw;
            }
            finally
            {
                Image
                    TargetImage = null;
                long
                    TargetImageSizeInByte = 0;

                int TargetQuality = 0;

                switch (BestImage)
                {
                    case 0:
                        TargetImage = ImageLo;
                        TargetImageSizeInByte = SizeInByteLo;
                        TargetQuality = MinQuality;
                        break;
                    case 1:
                        TargetImage = ImageMd;
                        TargetImageSizeInByte = SizeInByteMd;
                        TargetQuality = MidQuality;
                        break;
                    case 2:
                        TargetImage = ImageHi;
                        TargetImageSizeInByte = SizeInByteHi;
                        TargetQuality = MaxQuality;
                        break;
                }

                if (null != TargetImage && objImageExt.SizeInByte > TargetImageSizeInByte)
                {
                    System.Diagnostics.Trace.WriteLine(string.Format("[{0}][{1:00}][{2:00000}]: {3}", BestImage, TargetQuality, TargetImageSizeInByte, objImageExt.FileName));
                    string FileName = objImageExt.FileName;
                    objImageExt.Image.Dispose();
                    objImageExt = null;
                    TargetImage.Save(FileName);
                    TargetImage.Dispose();
                    TargetImage = null;
                }

                if (null != objImageExt)
                {
                    objImageExt.Image.Dispose();
                    objImageExt = null;
                }
            }
        }

#endif // (OLD_CODE)
        public void ScaleDownImage2FitMaxFileSize(string imageFileName, long maxFileSizeInByte)
        {
            ImageExt objImageExt = null;
            try
            {
                objImageExt = new ImageExt(imageFileName);

                //if (true == KeyDays.Lib.Threading.UserWorkItem.IsTraceOn)
                //{
                //    MyDiagno.Trace.TraceInformation(string.Format("[{0}]: {1}", objImageExt.FileName, objImageExt.SizeInByte));
                //}

                ImageResult objImageResult = this.ScaleDownImage2FitMaxFileSize(objImageExt, maxFileSizeInByte, this.Quality, 99, null, null);

                if (null != objImageResult && objImageExt.SizeInByte > objImageResult._ImageSizeInByte)
                {
                    if (true == KeyDays.Lib.Threading.UserWorkItem.IsTraceOn)
                    {
                        MyDiagno.Trace.TraceInformation(string.Format("[{0:00}][{1:00000}]: {2}", objImageResult._Quality, objImageResult._ImageSizeInByte, objImageExt.FileName));
                    }
                    string FileName = objImageExt.FileName;
                    objImageExt.Image.Dispose();
                    objImageExt = null;
                    objImageResult._Image.Save(FileName);
                    //objImageResult._Image.Save(@"C:\Temp\ImageTest-001\New.JPG");
                    objImageResult._Image.Dispose();
                    objImageResult._Image = null;
                }

                if (null != objImageExt)
                {
                    objImageExt.Image.Dispose();
                    objImageExt = null;
                }
            }
            catch (Exception exp)
            {
                MyDiagno.Trace.TraceError(exp, SeverityLevels.High);
                if (null != objImageExt)
                {
                    objImageExt.Image.Dispose();
                    objImageExt = null;
                }
                throw;
            }
        }
        private ImageResult ScaleDownImage2FitMaxFileSize(ImageExt objImageExt, long maxFileSizeInByte, int minQuality, int maxQuality, ImageResult maxResults, ImageResult minResults)
        {

            //if (true == KeyDays.Lib.Threading.UserWorkItem.IsTraceOn)
            //{
            //    MyDiagno.Trace.TraceInformation(string.Format("minQuality={0}, maxQuality={1}", minQuality, maxQuality));
            //    if (null != maxResults)
            //    {
            //        MyDiagno.Trace.TraceInformation(string.Format("maxResults:  Difference={0}, ImageSizeInByte={1}, Quality={2}", maxResults._Difference, maxResults._ImageSizeInByte, maxResults._Quality));
            //    }
            //    if (null != minResults)
            //    {
            //        MyDiagno.Trace.TraceInformation(string.Format("minResults:  Difference={0}, ImageSizeInByte={1}, Quality={2}", minResults._Difference, minResults._ImageSizeInByte, minResults._Quality));
            //    }
            //}

            // if (0 == System.Math.Abs(maxFileSizeInByte - objImageExt.SizeInByte) / 1024)
            if (objImageExt.SizeInByte <= maxFileSizeInByte)
            {
                maxResults = new ImageResult();
                maxResults._Image = (Image)objImageExt.Image.Clone();
                maxResults._ImageSizeInByte = objImageExt.SizeInByte;
                maxResults._Quality = maxQuality;
                maxResults._Difference = System.Math.Abs(maxFileSizeInByte - maxResults._ImageSizeInByte);

                return maxResults;
            }

            //--- Let us see what is happening at the central point
            ImageResult MidResults = new ImageResult();
            MidResults._Quality = ((null == maxResults ? maxQuality : maxResults._Quality) + (null == minResults ? minQuality : minResults._Quality)) / 2;
            MidResults._Image = this.ScaleDownImage(objImageExt, this.MaxSize, MidResults._Quality, out MidResults._ImageSizeInByte);
            MidResults._Difference = System.Math.Abs(maxFileSizeInByte - MidResults._ImageSizeInByte);

            //if (0 == MidResults._Difference / 1024)
            //{
            //    return MidResults;
            //}

            if (MidResults._ImageSizeInByte < maxFileSizeInByte)
            {
                if (MidResults._Quality + 1 == (null == maxResults ? maxQuality : maxResults._Quality))
                {
                    // return null == maxResults ? MidResults : MidResults._Difference < maxResults._Difference ? MidResults : maxResults;
                    return MidResults;
                }
                return this.ScaleDownImage2FitMaxFileSize(objImageExt, maxFileSizeInByte, MidResults._Quality, (null == maxResults ? maxQuality : maxResults._Quality), maxResults, MidResults);
            }

            if (null == maxResults)
            {
                maxResults = new ImageResult();
                maxResults._Quality = maxQuality;
                maxResults._Image = this.ScaleDownImage(objImageExt, this.MaxSize, maxResults._Quality, out maxResults._ImageSizeInByte);
                maxResults._Difference = System.Math.Abs(maxFileSizeInByte - maxResults._ImageSizeInByte);

                //if (0 == maxResults._Difference / 1024)
                //{
                //    return maxResults;
                //}
            }

            if (maxResults._Quality < minQuality)
            {
                return maxResults;
            }

            if (null == minResults)
            {
                minResults = new ImageResult();
                minResults._Quality = minQuality;
                minResults._Image = this.ScaleDownImage(objImageExt, this.MaxSize, minResults._Quality, out minResults._ImageSizeInByte);
                minResults._Difference = System.Math.Abs(maxFileSizeInByte - minResults._ImageSizeInByte);

                //if (0 == minResults._Difference / 1024)
                //{
                //    return minResults;
                //}
            }

            if (MidResults._Quality - 1 == (null == minResults ? minQuality : minResults._Quality))
            {
                // return null == minResults ? MidResults : MidResults._Difference < minResults._Difference ? MidResults : minResults;
                return null == minResults ? MidResults : (MidResults._ImageSizeInByte < minResults._ImageSizeInByte ? MidResults : minResults);
            }

            return this.ScaleDownImage2FitMaxFileSize(objImageExt, maxFileSizeInByte, minResults._Quality, MidResults._Quality, MidResults, minResults);
        }

        public Size CalculateScaleDownImageSize(Size originalSize, Size maxSize)
        {
            return this.CalculateScaleDownImageSize(originalSize, maxSize, ImageSizeAdjustmentTypes.Auto);
        }
        public Size CalculateScaleDownImageSize(Size originalSize, Size maxSize, ImageSizeAdjustmentTypes imageSizeAdjustmentType)
        {
            return ImageManager.CalculateScaleDownImageSize(originalSize, maxSize, imageSizeAdjustmentType, this.Strech);
        }
        public static Size CalculateScaleDownImageSize(Size originalSize, Size maxSize, ImageSizeAdjustmentTypes imageSizeAdjustmentType, bool strech)
        {
            //--- If the original size same as the max size then we are done.
            if (originalSize == maxSize)
            {
                return originalSize;
            }

            SizeF AdjustedSize = new SizeF();

            if (false == strech)
            {
                float Multiplier = (float)maxSize.Width / originalSize.Width;

                if ((originalSize.Height * Multiplier) <= (float)maxSize.Height)
                {
                    AdjustedSize.Height = Multiplier * (float)originalSize.Height;
                    AdjustedSize.Width = maxSize.Width;
                    return Size.Truncate(AdjustedSize);
                }

                Multiplier = (float)maxSize.Height / originalSize.Height;
                AdjustedSize.Height = maxSize.Height;
                AdjustedSize.Width = Multiplier * (float)originalSize.Width;

                return Size.Truncate(AdjustedSize);
            }

            //--- The aspect ratio of an image is its width divided by its height.
            float AspectRatio = (float)maxSize.Width / (float)maxSize.Height;

            if (ImageSizeAdjustmentTypes.Auto == imageSizeAdjustmentType)
            {
                imageSizeAdjustmentType = originalSize.Height < originalSize.Width ? ImageSizeAdjustmentTypes.Height : ImageSizeAdjustmentTypes.Width;
            }

            //--- We have been asked to adjust Width, that means we need to keep the Height constant
            if (ImageSizeAdjustmentTypes.Width == imageSizeAdjustmentType)
            {
                //--- Ok we need to Height costant but it shout not exceed MaxHeight
                AdjustedSize.Height = System.Math.Min(originalSize.Height, maxSize.Height);
                //--- Now adjust the Width as the optimized Height
                AdjustedSize.Width = true == strech ? AdjustedSize.Height * AspectRatio : System.Math.Min(originalSize.Width, AdjustedSize.Height * AspectRatio);
                return Size.Truncate(AdjustedSize);
            }

            //--- We have been asked to adjust Height, that means we need to keep the Width constant
            //--- Ok we need to Width costant but it shout not exceed MaxWidth
            AdjustedSize.Width = System.Math.Min(originalSize.Width, maxSize.Width);
            //--- Now adjust the Width as the optimized Height
            AdjustedSize.Height = true == strech ? AdjustedSize.Height * AspectRatio : System.Math.Min(originalSize.Height, AdjustedSize.Width / AspectRatio);

            return Size.Truncate(AdjustedSize);
        }
        /// <summary>
        /// CDPFs cannot handle progressive JPEG images so, this method validates whether an image is valid and 
        /// converts progressive JPEGs to non-progressive JPEGs.
        /// Assumptions:
        /// 1. This method will only be called for images that are less than the threshold size
        /// 2. The image file must already exist, since this method doesn't verify that
        /// </summary>
        /// <param name="sourceImageFileName">The full path of the image to be validated and converted</param>
        /// <returns>Full path of verified and/or converted .jpg file name</returns>
        public static string ValidateAndConvertImage(string sourceImageFileName)
        {
            FileStream _imageStream = new FileStream(sourceImageFileName, FileMode.Open, FileAccess.Read);
            string _filePath = Path.GetDirectoryName(sourceImageFileName);
            if (!_filePath.EndsWith(@"\"))
            {
                _filePath += @"\";
            }
            string _fileName = Path.GetFileNameWithoutExtension(sourceImageFileName);
            string _tempTargetImageFilePath = string.Format("{0}{1}.jpg", _filePath, "temp." + System.Guid.NewGuid().ToString("N"));
            string _targetImageFilePath = string.Format("{0}{1}.jpg", _filePath, _fileName);
            ImageInfo imageInfo = new ImageInfo(_imageStream);
            //Check whether the image is valid
            if (true == imageInfo.Check())
            {
                //If image is not jpeg OR a progressive jpeg then converted it to a non-progressive jpeg
                if (0 != imageInfo.Format || true == imageInfo.Progressive)
                {
                    Bitmap _imageToConvert = new Bitmap(sourceImageFileName);
                    //Get the jpeg image encoder
                    ImageCodecInfo _jpegCodecInfo = ImageCodecInfo.GetImageEncoders().First(rec => rec.MimeType == "image/jpeg");
                    //Set the encoder parameters
                    EncoderParameters _jpegCodecParameters = new EncoderParameters(1);
                    //Since, the image size is already less than the threshold size, maintain the image quality such that the file size of the
                    //new image doesn't exceed that of the original one
                    //A quality level of 0 corresponds to the greatest compression, and a quality level of 100 corresponds to the least compression
                    _jpegCodecParameters.Param[0] = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, 100L);
                    //Save new non-progressive jpeg with a temporary filename
                    _imageToConvert.Save(_tempTargetImageFilePath, _jpegCodecInfo, _jpegCodecParameters);
                    //Dispose off the stream and the bitmap objects
                    _imageStream.Flush();
                    _imageStream.Dispose();
                    _imageStream = null;
                    //Disposing off the bitmap object is critical since, it releases the handle and allows the original image to be deleted and the new image to be renamed
                    _imageToConvert.Dispose();
                    _imageToConvert = null;
                    //Delete the source image and rename the temporary image as the original source image
                    File.Delete(sourceImageFileName);
                    File.Move(_tempTargetImageFilePath, _targetImageFilePath);
                }
                else
                {
                    //If the image is valid then perform clean-up
                    //Dispose off the stream
                    _imageStream.Close();
                    _imageStream.Dispose();

                    if (".jpg" != Path.GetExtension(sourceImageFileName))
                    {
                        File.Move(sourceImageFileName, sourceImageFileName.Replace(Path.GetExtension(sourceImageFileName), ".jpg"));
                    }
                }
            }
            else
            {
                //Dispose off the stream
                _imageStream.Close();
                _imageStream.Dispose();
                //If not a valid image then throw invalid image exception
                throw new InvalidImageFile(new FileInfo(sourceImageFileName));
            }

            return _targetImageFilePath;
        }
        public static ulong FilterOutUnNecessaryImages(string sourceImageFileName, ImageProperties imageProperties)
        {
            return ImageManager.FilterOutUnNecessaryImages(sourceImageFileName, imageProperties.Size, imageProperties.Dimensions.Height, imageProperties.Dimensions.Width);
        }
        /// <summary>
        /// Unnecessary images such as the mail signatures should not be sent to CDPFs since they won't be displayed properly
        /// and the users will unnecessarily have to pay for something they don't want.
        /// This method checks the image file size and the dimensions and returns bit mask based on the results
        /// </summary>
        /// <param name="sourceImageFileName">The full path of the image to be filtered</param>
        /// <param name="thresholdSize">Threshold file size in KB</param>
        /// <param name="thresholdHeightInPX">Threshold image height in pixels</param>
        /// <param name="thresholdWidthInPX">Threshold image width in pixels</param>
        public static ulong FilterOutUnNecessaryImages(string sourceImageFileName, double thresholdSize, int thresholdHeightInPX, int thresholdWidthInPX)
        {
            ulong _result = 0x0000;
            Bitmap _imageToFilter = new Bitmap(sourceImageFileName);
            int _imageHeight = _imageToFilter.Height;
            int _imageWidth = _imageToFilter.Width;
            //Dispose off the bitmap object to relinquish the handle so that the image can be deleted if the need be
            _imageToFilter.Dispose();
            _imageToFilter = null;
            FileInfo _fileInfo = new FileInfo(sourceImageFileName);
            double _fileSize = _fileInfo.Length / 1024;
            if (_fileSize != thresholdSize)
            {
                if (_fileSize > thresholdSize)
                {
                    _result |= FileSizeGT;
                }
                else
                {
                    _result |= FileSizeLT;
                }
            }
            if (_imageHeight != thresholdHeightInPX)
            {
                if (_imageHeight > thresholdHeightInPX)
                {
                    _result |= ImageHeightGT;
                }
                else
                {
                    _result |= ImageHeightLT;
                }
            }
            if (_imageWidth != thresholdWidthInPX)
            {
                if (_imageWidth > thresholdWidthInPX)
                {
                    _result |= ImageWidthGT;
                }
                else
                {
                    _result |= ImageWidthLT;
                }
            }
            return _result;
        }
        
        #endregion methods
    }
}
