﻿/*
    Copyright 2002-2014 KeyDays LLC.

    This file is a part of KeyDays.Lib.
    KeyDays.Lib is licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. 
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using System.IO;

using KeyDays.Lib;

namespace KeyDays.Lib.Cryptography
{
    public sealed class RijndaelExt : KeyDays.Lib.IDisposableExt
    {
        #region properties

        public static readonly char [] SpecialCharacters;
        public static readonly int PasswordMinLength;
        public static readonly string DefaultPassword;
        private static readonly string[] Salts;        
        public bool EnforcePasswordPolicy { get; private set; }
        public Rijndael Algorithm { get; private set; }
        public ICryptoTransform Encryptor { get; private set; }
        public ICryptoTransform Decryptor { get; private set; } 

        #endregion properties

        #region constructors
        
        static RijndaelExt()
        {
            RijndaelExt.SpecialCharacters = new char[] { '!', '@', '#', '$', '^', '&', '*', '-' };
            RijndaelExt.PasswordMinLength = 8;
            RijndaelExt.DefaultPassword = "31AFE8D4-C878-4b74-BBC0-EC00884A218C";

            RijndaelExt.Salts = new string[] 
            {
                "5108D31D-635B-4701-9EEB-1F40CA22A98E",
                "3DBCC25C-A633-4dd4-89F7-ADFF65BCB61E",
                "50B7F309-B49E-416b-A1B7-49DC3A51AC81",
                "6266D781-57E2-450a-9369-DA78E60D5F67",
                "B4ADD32B-D59D-40c7-BE8D-CA41F5FDE511",
                "BC2D57D3-24CD-45dd-BC1F-3BACD6CDFE0B",
                "F7B7B12F-E056-4ac1-B134-9D022108A597",
                "D9642780-BDA8-46c6-B0E6-34AC9B3A5C03",
                "6F5E950E-58DF-4fda-8773-3AACCC114C65",
                "E6A61DF4-1843-4c4a-B62B-79A6A7DEC825",
            };
        }

        public RijndaelExt()
            : this(RijndaelExt.DefaultPassword)
        { 
            //--- Do nothing.
        }

        public RijndaelExt(string password)
            : this(password, false)
        {
            //--- Do nothing.
        }

        public RijndaelExt(string password, bool enforcePasswordPolicy)
        {
            this.EnforcePasswordPolicy = enforcePasswordPolicy;
            password.extParameterValidation("password");

            char[] PasswordChars = password.ToCharArray();

            if (true == this.EnforcePasswordPolicy)
            {
                var SpecialChars = PasswordChars.Intersect(SpecialCharacters);
                var LowerCaseChars = PasswordChars.Where(rec => 'a' <= rec && 'z' >= rec);
                var UpperCaseChars = PasswordChars.Where(rec => 'A' <= rec && 'Z' >= rec);
                var Digits = PasswordChars.Where(rec => '0' <= rec && '9' >= rec);

                if (RijndaelExt.PasswordMinLength > password.Length || 0 == SpecialChars.Count() || 0 == LowerCaseChars.Count() || 0 == UpperCaseChars.Count() || 0 == Digits.Count())
                {
                    throw new KeyDays.Lib.Exceptions.PasswordPolicyException();
                }   
            }

#if(OLD_Cryptograhy)
            byte[] Salt = RijndaelExt.Salts[PasswordChars.Sum(rec => (int)rec) % RijndaelExt.Salts.Length].extStrToByteArray();

#else
            byte[] Salt = Encoding.UTF8.GetBytes(RijndaelExt.Salts[PasswordChars.Sum(rec => (int)rec) % RijndaelExt.Salts.Length]);
#endif//(OLD_Cryptograhy)
            this.Algorithm = Rijndael.Create();
            this.Algorithm.Padding = PaddingMode.ISO10126;
            //--- http://stackoverflow.com/questions/740837/how-to-create-a-password-protected-file-in-c
            Rfc2898DeriveBytes aRfc2898DeriveBytes = new Rfc2898DeriveBytes(password, Salt);
            // BlockSize, KeySize in bit --> divide by 8
            this.Algorithm.IV = aRfc2898DeriveBytes.GetBytes(this.Algorithm.BlockSize / 8);
            this.Algorithm.Key = aRfc2898DeriveBytes.GetBytes(this.Algorithm.KeySize / 8);

            this.Encryptor = this.Algorithm.CreateEncryptor();
            this.Decryptor = this.Algorithm.CreateDecryptor();
        }

        #endregion constructors

        #region methods

        //--- http://www.codeproject.com/KB/security/DotNetCrypto.aspx
        public string Encrypt(string clearText)
        {
            byte[] ClearBytes = System.Text.Encoding.Unicode.GetBytes(clearText);
            byte[] EncryptedData = this.Encrypt(ClearBytes);
            return Convert.ToBase64String(EncryptedData);
        }

        public byte[] Encrypt(byte[] clearData)
        {
            //--- Create a MemoryStream to accept the encrypted bytes 
            MemoryStream aMemoryStream = new MemoryStream();
            CryptoStream aCryptoStream = new CryptoStream(aMemoryStream, this.Encryptor, CryptoStreamMode.Write);

            //--- Write the data and make it do the encryption 
            aCryptoStream.Write(clearData, 0, clearData.Length);
            aCryptoStream.Close();

            //--- Now get the encrypted data from the MemoryStream.
            //--- Some people make a mistake of using GetBuffer() herewhich is not the right way. 
            byte[] encryptedData = aMemoryStream.ToArray();
            return encryptedData;
        }

        public string Decrypt(string cipherText)
        {
            //--- First we need to turn the input string into a byte array.  We presume that Base64 encoding was used 
            byte[] cipherBytes = Convert.FromBase64String(cipherText);

            byte[] decryptedData = this.Decrypt(cipherBytes);

            //--- Now we need to turn the resulting byte array into a string. A common mistake would be to use an Encoding class for that. It does not work because not all byte values can be represented by characters. We are going to be using Base64 encoding that is designed exactly for what we are trying to do. 
            return System.Text.Encoding.Unicode.GetString(decryptedData);
        }

        public byte[] Decrypt(byte[] cipherData)
        {
            //--- Create a MemoryStream that is going to accept the decrypted bytes 
            MemoryStream aMemoryStream  = new MemoryStream();

            //--- Create a CryptoStream through which we are going to be pumping our data. 
            //--- CryptoStreamMode.Write means that we are going to be writing data to the stream and the output will be written in the MemoryStream we have provided. 
            CryptoStream aCryptoStream = new CryptoStream(aMemoryStream, this.Decryptor, CryptoStreamMode.Write);

            //--- Write the data and make it do the decryption 
            aCryptoStream.Write(cipherData, 0, cipherData.Length);

            //--- Close the crypto stream (or do FlushFinalBlock). This will tell it that we have done our decryption and there is no more data coming in, and it is now a good time to remove the padding and finalize the decryption process. 
            aCryptoStream.Close();

            //--- Now get the decrypted data from the MemoryStream. Some people make a mistake of using GetBuffer() here, which is not the right way. 
            byte[] decryptedData = aMemoryStream.ToArray();
            return decryptedData;
        }

        #endregion methods

        #region IDisposable Members

        public bool IsDisposed { get; private set; }

        // protected virtual void Dispose(DisposalTypes disposalType)
        private void Dispose(DisposalTypes disposalType)
        {
            if (true == this.IsDisposed)
            {
                return;
            }

            if (DisposalTypes.External == disposalType)
            {
                //--- Clean-up of managed code.
                if (null != this.Encryptor)
                {
                    this.Encryptor.Dispose();
                    this.Encryptor = null;
                }

                if (null != this.Decryptor)
                {
                    this.Decryptor.Dispose();
                    this.Decryptor = null;
                }
            }
            //--- Clean-up of unmanaged code.

            //--- Mark the object as disposed.
            this.IsDisposed = true;
        }

        public void Dispose()
        {
            this.Dispose(DisposalTypes.External);
            GC.SuppressFinalize(this);
        }

        ~RijndaelExt()
        {
            this.Dispose(DisposalTypes.Internal);
        }

        #endregion IDisposable
    }
}
